@@ -5,8 +5,10 @@
  *                           Laboratoire MASI, Institut Blaise Pascal
  *                           Universite Pierre et Marie Curie (Paris VI)
  *
- * This file can be redistributed under the terms of the GNU Library General
- * Public License
+ * %Begin-Header%
+ * This file may be redistributed under the terms of the GNU Library
+ * General Public License, version 2.
+ * %End-Header%
  */
 
 /*
@@ -14,9 +16,14 @@
  * 93/10/30	- Creation
  */
 
+#ifndef _LARGEFILE_SOURCE
 #define _LARGEFILE_SOURCE
+#endif
+#ifndef _LARGEFILE64_SOURCE
 #define _LARGEFILE64_SOURCE
+#endif
 
+#include "config.h"
 #if HAVE_ERRNO_H
 #include <errno.h>
 #endif
@@ -32,16 +39,19 @@
 
 #include "e2p.h"
 
-#ifdef O_LARGEFILE
-#define OPEN_FLAGS (O_RDONLY|O_NONBLOCK|O_LARGEFILE)
-#else
-#define OPEN_FLAGS (O_RDONLY|O_NONBLOCK)
+#ifndef O_LARGEFILE
+#define O_LARGEFILE 0
 #endif
+#ifndef O_NOFOLLOW
+#define O_NOFOLLOW 0
+#endif
+
+#define OPEN_FLAGS (O_RDONLY|O_NONBLOCK|O_LARGEFILE|O_NOFOLLOW)
 
 int fgetflags (const char * name, unsigned long * flags)
 {
 	struct stat buf;
-#if HAVE_STAT_FLAGS
+#if HAVE_STAT_FLAGS && !(APPLE_DARWIN && HAVE_EXT2_IOCTLS)
 
 	if (stat (name, &buf) == -1)
 		return -1;
@@ -61,28 +71,35 @@ int fgetflags (const char * name, unsigned long * flags)
 #endif
 
 	return 0;
-#else
-#if HAVE_EXT2_IOCTLS
+#elif APPLE_DARWIN && HAVE_EXT2_IOCTLS
+	int f, save_errno = 0;
+
+	f = -1;
+	save_errno = syscall(SYS_fsctl, name, EXT2_IOC_GETFLAGS, &f, 0);
+	*flags = f;
+	return (save_errno);
+#elif HAVE_EXT2_IOCTLS
 	int fd, r, f, save_errno = 0;
 
-	if (!stat(name, &buf) &&
-	    !S_ISREG(buf.st_mode) && !S_ISDIR(buf.st_mode)) {
-		goto notsupp;
-	}
-	fd = open (name, OPEN_FLAGS);
-	if (fd == -1)
+	fd = open(name, OPEN_FLAGS);
+	if (fd == -1) {
+		if (errno == ELOOP || errno == ENXIO)
+			errno = EOPNOTSUPP;
 		return -1;
-	r = ioctl (fd, EXT2_IOC_GETFLAGS, &f);
-	if (r == -1)
+	}
+	r = ioctl(fd, EXT2_IOC_GETFLAGS, &f);
+	if (r == -1) {
+		if (errno == ENOTTY)
+			errno = EOPNOTSUPP;
 		save_errno = errno;
+	}
 	*flags = f;
-	close (fd);
+	close(fd);
 	if (save_errno)
 		errno = save_errno;
 	return r;
-#endif /* HAVE_EXT2_IOCTLS */
-#endif
-notsupp:
+#else
 	errno = EOPNOTSUPP;
 	return -1;
+#endif
 }
